import UIKit

/* Protocols */


// A blueprint of methods, properties, operators, subscripts.
// It can be adopted by class, struct, enum

protocol WalkableProtocol {
    func walk()
}

class Dog : WalkableProtocol {
    func walk() {}   // If Dog doesn't have walk(), compiler fails
}
// WalkableProtocol is adopted by Dog
// Dog conforms to WalkableProtocol


struct Goose :WalkableProtocol {
    func walk() {}
}

enum Fox : WalkableProtocol {
    func walk() { }
}

let x = Goose()
let w:WalkableProtocol = x  // Upcasting
let x2:Goose = w as! Goose   // Downcasting

// Summary of cast
// 1. Cast between super classes and sub-classes  -- class
// 2. Cast to/from AnyObject                      -- class
// 3. Cast to/from Any                            -- class, struct, enum, closure
// 4. Cast to/from a prototocol, or between protocols -- class, struct, enum








// Protocol of Properties
protocol NamedProtocol {
    var name: String {get set}  // name can be read and written
}


class Cat: WalkableProtocol, NamedProtocol {
    func walk() {}
    var name:String = ""  // let name:String = "" --> Error
}


protocol AgedProtocol {
    var age: Int {get}   // age can be read
}

class A : AgedProtocol {
    let age: Int = 5   // age is readable and writable.
}

//class B : AgedProtocol {
//    var age : Int  {   // Computed property, readonly.
//        return 2 * 3
//    }
//}








// Mutating Methods
protocol Togglable {
    mutating func toggle()
}

struct C : Togglable{   // struct and enum have mutating methods
    mutating func toggle() { /* Modifying itself's state */ }
}

class D : Togglable {
    func toggle() { /* Modifying itself's state */ }
}


// Protocols as Types
var a = [WalkableProtocol]()
a.append(Dog())
a.append(Cat())

func foo(animal: WalkableProtocol) {}  // foo can talk either Dog or Cat
foo(animal: Dog())












// Protocol inheritance
protocol RunableProtocol : WalkableProtocol {
    func run()
}
// It requires 2 methods: walk() and eat()

class Wolf : RunableProtocol {
    func walk() { }
    func run() { }
}


// Protocol Composition
var y : WalkableProtocol & NamedProtocol

protocol habitableProtocol {
    var creatures: [WalkableProtocol & NamedProtocol] {get}
}












// Proctocol Conformance Check and Cast

@objc protocol Edible { }

class Apple : Edible { }

let a1 = Apple()
let e:Edible = a1 as Edible   // upcasting
let e2:Edible = a1    // also works

if e is Apple {
    print("e is an Apple")
}

// Downcasting
let a2 = e as! Apple   // runtime error  if e is not an Apple
let a3 = e as? Apple  // returns nil if e is not an Apple


//  @objc means the protocol is exposed to Objective-C code, which
//  also means the protocol can only be used on class.













// Optional Protocol Requirements
@objc protocol Seenable {
    
    var color:String {get set}
    @objc optional func greet() -> String // Optional for comforming class to implement
}

class Kite : Seenable {
    var color:String = ""
    func greet() -> String {
        return "Dummy"
    }
}

let k : Seenable = Kite()
k.greet!()   // Runtime error if Kite doesn't have greet()

k.greet?()   // if greet() is not implemented by Kite, returns nil
// It can be used in optional chaining and optional binding
























