
// Access Control

// Module:  App or Framework
//    Can be imported by another module
//    Each Xcode build target is a module
import UIKit


// Public Access:  accessible outside the module
// Internal Access:  accessible inside the module
// Private Access:  accessible inside the same source file


/*
 * 1. Date Type
 */
public class MyPublicClass {    // Public class
    
    public var name:String?     // Public property
    
    internal let height:Double = 0  // Explicitly internel
    
    var age:Int?                 // Implicitly internal -- default access level
    
    private func foo() {}   // Private method
}


class MyInternalClass {  // Default access level -- 'internal' is optional
    internal let height:Double = 0  // Explicitly internel
    var age:Int?               // Implicitly internal -- default access level
    private func foo() {}      // private
    //public var name:String?    // Illegal
}


private class MyPrivateClass {
    private func foo() {}      // explicitly private
    var age:Int?              // Implicitly private  -- default access level
    //internal let height:Double = 0  // Illegal
    //public var name:String?    // Illegal
}

// These rules also apply to compiler generated initializers









/*
 * 2. Data cannot have higher access level than its type
 */

// A property cannot be more public than its type
public class MyClass {
    var m0:MyInternalClass?  // OK
//    var m1:MyPrivateClass?   // m1 cannot be internal
//    public let m2:MyInternalClass?  // m2 cannot be public
//    let m3 = (23, MyPrivateClass())  // m3 cannot be internal
}


//
private func foo(x:MyInternalClass)->MyPrivateClass {return MyPrivateClass()}
// foo() must be a private function because its signature has a private type



// default memberwise initializer have the lowest access level of its properties
public struct MyStruct {
    public var m1:Int = 2
    var m2: String = ""
    private var m3:Double = 3
}















/*
 * 3. Enumeration
 */
// Individual members automatically receive the same access level as the enumeration
public enum Direction {
    case NORTH,SOUTH,EAST,WEST   // They are all public
}

/*
 * 4. Nested Type
 */
// Nested types defined within a private type have an default access level of
// private. 
// Nested types defined within a public type or an internal type have an
// default access level of internal













/*
 * 5. Subclassing Type
 */
// A subclass cannot have a higher access level than its superclass
class A {}
private class B: A { }  // B and B2 cannot be public
internal class B2: A { }

// An override can make an inherited class member more accessible than its superclass version
class C {
    private func foo() {}
}

class D : C {
    override internal func foo() {}  // Can we make foo() public?
}

// Subclass could access superclass's private member, if they are in same source file.
class E : C {
    func goo() {
        super.foo()   // This is illegal in C++ or Objective-C
    }
}







/*
 * 6. Getters and Setters
 */
// Getters and setters by default have same access level as their properties

// Change the defaults:
class F {
    private(set) var m0 = 2.3
    
    private(set) var m1:Int = 0
    // m1's getter has internal access, setter has private access
}



















/*
 * 7. Protocols
 */
// Protocol members have the same access level as protocol itself.
public protocol Walkable {
    func walk()    // walk() is also public
}

private protocol Runable {
    func run()    // run() is also private
}

// Protocol cannot have higher access level than the one it inherits from
private protocol Goo : Runable {  // Goo cannot be internal or public
    func goo()
}











